/*  initialise.c - This program is used to create a base 'net' file
    with only the initial letters being present. 
    Also handle place for defrag() - 'net' file defragmenter
    
    responce_node()
    initialise()
    defrag()     
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "datastru.h"

/**************************************************************************
 * responce_node(node)
 * returns TRUE or FALSE depending on whether node is a responce char that
 * should trigger a reply in chat-mode.
 */
 
int responce_node(int node)
{
  switch(node)
  {
    case (int)'.':
    case (int)'?':
    case (int)'!':
    case (int)':':
    case (int)';': return(TRUE);
    
    default: return(FALSE);
  }
}

/************************************************************************
 * initialise()  - set up a 'net' file.
 */
 
void initialise()
{
  net_type *net;
  int loop;
  int fp;

#ifdef DOS
  net=(net_type *)malloc(sizeof(net_type));
#else   /* We're in UNIX mode... */
  fp=open("net", O_RDWR | O_CREAT | O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if (fp<0) {
    perror("open('net')");
    exit(1);
  }
  lseek(fp,sizeof(net_type)-1,SEEK_SET); 
  write(fp, "", 1); 
  net=(net_type *)mmap(0,sizeof(net_type),PROT_READ+PROT_WRITE,MAP_SHARED,fp,0);
  if (net<=0)
  {
    printf("WOA! the 'net' datastructure is too big!!! lower the\n");
    printf("values in datastructs.h please! \n");
    exit(0);
  }
#endif

  net->ptr=MAX_DATA_SIZE;  /* top of net->data*/
  net->mode=0;
  net->nodespace=0; net->actual_num_nodes=0;
  net->deleted_node=0;
  net->spoken=0;
  net->next_dream=NUM_BASE_NODES*2;
  for (loop=0; loop<MAX_IPOP; loop++) net->free[loop]=0;
  
  /* OK, all we have to do now is allocate the first NUM_BASE_NODES nodes
     as base nodes for I/O.  */
  for(loop=1; loop<=NUM_BASE_NODES; loop++)
  {
    new_node(net);
    /* set the responce bit on the spec-int... */
    if (responce_node(loop)) net->data[loop*TOP]=HIGH_BIT;
    /* set the first node, and the level to 1 */
    net->data[loop*TOP]+=loop*TOP+LEVEL_DIV;
  }
  
  /*lastly, write the network out to disk. */
#ifdef DOS
  save(net);

#else
  msync((void *)net,sizeof(net_type), MS_SYNC);
  munmap((void *)net,sizeof(net_type));
  close(fp);
#endif
}

/****************************************************************************
 * round_it_up()
 * Little routine used by defrag() to round up chunk sizes to standards.
 */
 
int round_it_up(int size)
{
   if (size==0) return(0);
   if (size==1) return(1);
   if (size==2) return(2);
   if (size<=4) return(4);
   if (size<=8) return(8);
   if (size<=16) return(16);
   if (size<=32) return(32);
   if (size<=64) return(64);
   if (size<=128) return(128);
   if (size<=256) return(256);
   if (size<=512) return(512);
   if (size<=1024) return(1024);
   return(((size-1)/1024)*1024+1024);
}

/*****************************************************************************
 * add_to_table()
 * used by defrag() when building a table of chunks
 * NOTE!!!! Table is kept in REVERSE ORDER, highest item first.
 */
 
void add_to_table(table_type *table,int posn,int node)
{

  int bot,mid,top, finished;
  int below, above;
  int loop;

  mid=0;

  /* printf("add_to_table() -item=%i, tabsize=%i\n",posn,table->size); */
  
  if (posn==0) return;
  /* firstly, work out where in the table that 'posn' fits... */
  if (table->size==0) mid=1;
  else if (posn>table->posn[1]) mid=1;
  else if (posn<table->posn[table->size]) mid=table->size+1;
  else
  {
    bot=1; top=table->size; finished=FALSE;
    while (bot<=top && !finished)
    {
      mid=bot+(top-bot)/2;
      if (mid==1) above=posn+1; else above=table->posn[mid-1];
      below=table->posn[mid]; 
      /* printf("posn=%i, above=%i, below=%i\n",posn,above,below); */
      if (posn>above) top=mid-1;
      else if (posn<below) bot=mid+1;
      else finished=TRUE;
    }
    if (!finished) {printf("WOA! add_to_table broken!\n"); exit(0); }
  }
 
  /* now move the right number of items up to make space, and copy new node in */
  for (loop=table->size; loop>=mid; loop--)
     {table->posn[loop+1]=table->posn[loop];
      table->node[loop+1]=table->node[loop]; }
  table->posn[mid]=posn; table->node[mid]=node;
  if (table->size<FRAG_SIZE) table->size++;
}

/*****************************************************************************
 * defrag(net)
 * This scary routine will shuffle around a net structure on disk,
 * defragmenting it as it goes...
 */
 
void defrag(net_type *net)
{
  table_type *table;
  int top,floor,ceiling;
  int posn;
  int loop;
  int gap, side, node, size,chunksize;
  int copy;
  int percent, old_percent;
   
  if (net->trace>=1)
    printf("****** Beginning Defragmentation.... ******\n");
    
  table=(table_type *)malloc(sizeof(table_type));
  table->size=0; 
  ceiling=MAX_DATA_SIZE-1; floor=0; percent=0; old_percent=0;
  
   /* OK, we build up the table N times... */
  while(ceiling>=net->ptr)
  {        
    /* go through every node, filling out table[]... */
    table->size=0;
    for (loop=0; loop<=net->nodespace; loop+=TOP)
    {
       /* firstly, LHS_OP... */
       posn=net->data[loop+LHS_OP];
       if (posn>ceiling);  
       else if (table->size==FRAG_SIZE-1 && posn<table->posn[table->size]);
       else add_to_table(table,posn,loop);   /* insert into the table */
       /* not RHS_OP... */
       posn=net->data[loop+RHS_OP];
       if (posn>ceiling);
       else if (table->size==FRAG_SIZE && posn<table->posn[table->size]);
       else add_to_table(table,posn,loop);  /* insert into the table */
    }
    
    ceiling=table->posn[1];
    percent=(100*(MAX_DATA_SIZE-ceiling)/(MAX_DATA_SIZE-net->ptr));
    if (percent>old_percent)
    {
      old_percent=percent;
      if (net->trace>=1) printf("*** %i percent complete...\n",percent);
    }
    if (table->size==1) /* i.e. we're finished now... */
    {
      if (net->trace>=1)
        printf("*** Finished defrag.  net->ptr was %i, now %i, saved %i. **\n",
              net->ptr, ceiling, ceiling-net->ptr);
     
      net->ptr=table->posn[1];  /*update the ceiling */
      /* lastly, blank out the net->free[] strctures... */
      for (loop=0; loop<MAX_IPOP; loop++) net->free[loop]=0;
#ifdef DOS
      save(net);
#else
      msync((void *)net,sizeof(net_type), MS_ASYNC); /*write changes back*/
#endif
      free(table);
      /*
      return(ceiling-net->ptr);
      */
      return;
    }
    
    /* OK, now that we've build a table, shuffle every item present to the
       top... */
    for (loop=2; loop<=table->size; loop++)
    {
       node=table->node[loop]; posn=table->posn[loop]; top=table->posn[loop-1];
       gap=top-posn; 
       side=0; /* LHS */
       if (net->data[node+RHS_OP]==posn) side=1;
       if (side==0) size=net->data[node+NUM_LHS_OP];
               else size=net->data[node+NUM_RHS_OP];
       chunksize=round_it_up(size);
       
       if (chunksize!=gap)  /* i.e. there's free space needing cleared up... */
       {
         /* firstly, copy everything upwards as necessary... */
         for (copy=size-1; copy>=0; copy--)
           net->data[top-chunksize+copy]=net->data[posn+copy]; 
         if (side==0) net->data[node+LHS_OP]=top-chunksize; 
                 else net->data[node+RHS_OP]=top-chunksize;
         table->posn[loop]=top-chunksize;
       }
    }
    /* re-adjust ceiling for next pass... */
    ceiling=table->posn[table->size];
  }
}

