/* meme machine - domain modelling using recursive information storage.
 * Routines defined here:
 *
 * off_the_end()
 * need_input()
 * compare()
 * analyse_match()
 * find_match()
 * find_greedy_match()
 * netlisten()
 * main()
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include "datastru.h"

#define STR_LEN 1024
char * home=NULL;
char * chatlog=NULL;


/*************************************************************************
 * off_the_end() - routine to handle running out of input
 * (This is a very infrequently called routine, so can be slow as a slug)
 */
int off_the_end(net_type *net, int node, nodelist_type *input, int posn)
{
  FILE *fp;

  /*printf("\n** OFF THE END!! node='%c' mode=%i\n",node/TOP,net->mode);
    printf("spoken=%i\n",net->spoken); */
  if (net->mode==1) return(posn);  /*we're in listen mode */
  /* next line ensures that we get to the beginning of a responce
     before beginning to speak back.... */
  else if (net->mode==2 &&  net->data[net->best]>=HIGH_BIT) net->mode=3; 
  /* deal with hitting the end of computer's talk... */
  else if (net->mode==3 && responce_node(node/TOP)) 
  { input->size++; input->data[input->size]=node;
     printf("%c",node/TOP); net->mode=4; }
	
  if (net->mode==2)  /*we're in scan forward to next responce mode */
  {
    /* printf("\nSKIP OVER '%c'\n",node/TOP); */
    /* input->size++; input->data[input->size]=node; */
    return(posn+1);
  }
  else if (net->mode==3) /* we're in speak forward to end of responce mode*/
  { 
     input->size++; input->data[input->size]=node; net->spoken++;
     printf("%c",node/TOP);
     fp=fopen(chatlog,"a"); fputc(node/TOP,fp); fclose(fp);
     return(posn+1);
  }
  else if (net->mode==4)  /*we've switched into prompt-mode */
  {
      need_input(net,input);
      /* switch back to listen mode... */
      net->mode=2; net->spoken=0;
      if (node==input->data[posn]) return(posn+1); else return(FALSE);
  }
  if (responce_node(node/TOP) && net->mode==2) net->mode=3;
  /* if (net->mode==3 && !responce_node(node/TOP)) net->mode=2; */
  /* functionality change! -kees */
  fprintf(stderr,"\nooh: returning FALSE.  Is that okay? -kees\n");
  return(FALSE);
}

/*************************************************************************
 * need_input()
 *  ;) another chat-bot interface function. It is called once a reply has
 * been sent to the user and more input is needed.
 * it prompts for a reply from stdin, which it also adds
 * to the end of the input stream. This allows find_match to recurse
 * further.
 *
 */
void need_input(net_type *net, nodelist_type *input)
{
  int start,loop;
  FILE *fp;
  FILE *botfp;
  char str[20];
  int bot_pid;
  int log_posn;
  int received_signal();
  char botpid[STR_LEN];
  int fgetcret;
  static int firsttime=1;

  /* set a clean return value */
  fgetcret=0;

  /* build our future filenames */
  snprintf(botpid,STR_LEN,"%s/.meme_bot.pid",home);

  if (!(fp=fopen(chatlog,"a"))) {
	fprintf(stderr,"fopen('%s')",chatlog);
	perror("");
	exit(1);
  }
 
  /* firstly, send reply out to both input[] and stdout */
  /* add_to_input(net,input,reply); */
  start=input->size+1;
  
  /* sanity check... are we in read mode? */
  if (net->mode==1) return;
  /* Deal with when the computer has nothing to say... */
  if (!firsttime) {
    /* but only after being spoken to */
    if (net->spoken<=1)
      {
	printf("hmmmm");
	fprintf(fp,"hmmmm");
      }        
    /* add return char to end of chat log... */
    fputc('\n',fp);
    printf("\n");
  }
  firsttime=0;
  fprintf(stderr,"---------------------------------------------\n");
  
  /* firstly, if we're talking to a bot, tell it that it has input and wait for
     human responce */
  if ((botfp=fopen(botpid,"r"))) {    
     fgets(str,10,botfp); fclose(botfp); /* read in the bot pid */
     
     bot_pid=atoi(str);
     fprintf(stderr,"BOT PID IS %i\n",bot_pid);
     signal(42,(void *)received_signal);
     /* Firstly, save off the current chatlog size.. */
     if (fseek(fp,0,SEEK_END)) {
	fprintf(stderr,"fseek('%s')",chatlog);
	perror("");
	exit(1);
     }
     log_posn=ftell(fp);
     fprintf(stderr,"OLD LOG_POSN=%i\n",log_posn);
     fprintf(stderr,"SENDING BOT a 42...\n");
     kill(bot_pid,42);  /*tell the bot to seek a reply */
     pause();  /* wait for the reply, then read out the reply*/
 
     if (fseek(fp,log_posn,SEEK_SET)) {
	fprintf(stderr,"fseek('%s')",chatlog);
	perror("");
        exit(1);
     }
     do
     {
       input->size++;
       input->data[input->size]=fgetc(fp);
       fprintf(stderr,"%c",input->data[input->size]);
       if (responce_node(input->data[input->size-1]) &&
          input->data[input->size]==32)
	  input->size--;
     } while (input->data[input->size]!=10);
     fprintf(stderr,"\n");
  }
  else /* read input from STDIN... */
  {
    do
    {
      input->size++;
      fgetcret=fgetc(stdin);
      if (fgetcret!=EOF) {
      	input->data[input->size]=(char)fgetcret;
      	if (responce_node(input->data[input->size-1]) &&
            input->data[input->size]==32) input->size--;
      }
    } while (fgetcret!=EOF && input->data[input->size]!='\n'); 
  }

  /* close chat log */
  fclose(fp);

  if (fgetcret!=EOF) {
     input->size--;
     if (!responce_node(input->data[input->size]))
     {
       input->size++;
       input->data[input->size]=(int)'.';
       input->data[input->size]=input->data[input->size];
     }
  }

  /* ugly, but who cares, it works */
  if (fgetcret==EOF || (input->size>2 && 
	 input->data[input->size-3]=='b' &&
         input->data[input->size-2]=='y' &&
         input->data[input->size-1]=='e' ))
  {
#ifdef DOS
     save(net);
#else
     if (msync((void *)net,sizeof(net_type), MS_SYNC)) {
	perror("msync('net')");
	exit(1);
     }
#endif
     fprintf(stderr,"bye.\n");
     exit(0);
  }
  if (net->trace>=1)
  {
    fprintf(stderr,"NEW INPUT='");
    for (loop=start; loop<=input->size; loop++) {
      fprintf(stderr,"%c",input->data[loop]);
    }
    fprintf(stderr,"'\n");
  }
  /* need to multiply the input by TOP for consistency now.... */
  for (loop=start; loop<=input->size; loop++)
     input->data[loop]=input->data[loop]*TOP;
  fprintf(stderr,">>");  /* prepare for the responce */
  fflush(NULL);
}

/****************************************************************************
 * received_signal() - do nothing 
 */
void received_signal(int sig)
{
}

/****************************************************************************
 * compare() - this recursive routine compares 'node' with input->data[posn].
 * If 'node' is not a base node, compare() recurses.
 * returns updated 'posn' if exact match found, otherwise returns 0/FALSE.
 * it no prior tweak is set, we will try one.
 * Now including support for a single tweak.
 */
 
int compare(net_type *net, int node, nodelist_type *input, int posn)
{

  int lhs,rhs;  /*lhs and rhs of 'node' if 'node' is not a base-node*/
  int result;
  int found;
  int first;
  int marker;
  int tracer;
  
  if (net->trace>=5)
  {
    printf("\ncompare(%i '",node); print_node(net,node); printf("' , '"); 
     for (tracer=0; tracer<5; tracer++)
        print_node(net,input->data[posn+tracer]);
     printf("')\n");
  }
  
  if (node>BASE_NODE_SPACE)  /*not looking at base nodes...*/
  {
     /* first off, bomb out early if first item of node does'nt match*/
    first=net->data[node] & FIRST_MASK;
    /* printf("first='%c' %i\n",(char)(first/8),first/8); */
    result=first-input->data[posn];
    if (result!=0 && posn<=input->size) {net->result=result; return(FALSE); }
    
    marker=node+LHS_IP;
    lhs=net->data[marker]; marker++; rhs=net->data[marker];
     
    result=compare(net,lhs,input,posn);  /* look at the LHS */
    if (result==-1) return(-1);  /*run off the end*/
    if (result==0)  return(FALSE);  /*no match */  
    
    /* if we made it here, we found an exact match on LHS*/
     
    found=compare(net,rhs,input,result);  /*now try and match RHS now*/
    return(found);  /* return result */
  }
  
  /*-----------------------------------------------------*/
  else  /*just comparing 2 base nodes, easy!! */
  {
    /* firstly, deal with running out of input... */
    if (posn>input->size)
    {
       found=off_the_end(net,node,input,posn);
       return(found);
    }
    else if (node==input->data[posn]) return(posn+1);
    else
    {
      if (node>input->data[posn]) net->result=1;
                             else net->result=-1;
      return(FALSE);
    }
  }   
}

/**************************************************************************
 * analyse_match()
 * 
 * called by find_match() - see if a found match is the best so far, and
 * recurse.
 */
 
void analyse_match(net_type *net, int metanode,
		   nodelist_type *input, int result_posn)
{
   int level_total;

  /* firstly, look for the best greedy_node... */
  /* printf("this level is %i, best level is %i\n",
          get_level(net,metanode),net->bestlevel); */
   /* printf("This node='"); print_node(net,metanode); printf("', greedy='");
      print_node(net,net->g_best); printf("'\n"); */
  level_total=get_level(net,metanode); /* +(get_level(net,net->g_best)/2); */
  if (level_total>net->bestlevel) 
  {
    if (net->trace>=2)
     { printf("**** BEST='"); print_node(net,metanode);
       printf("', lhs_outputs=%i, rhs_outputs=%i\n",
              net->data[metanode+NUM_LHS_OP],net->data[metanode+NUM_RHS_OP]); }
    net->best=metanode;
    net->bestlevel=level_total;
    net->bestnext=result_posn;
    if (net->mode==2 && net->data[metanode]>=HIGH_BIT) net->mode=3;
  }
  find_match(net,metanode,input,result_posn);
}

/****************************************************************************
 * find_match()
 * try to find a best suitable metanode of 'node' that best matches
 * the current input, from posn onwards.
 * Return value is TRUE OR FALSE. if TRUE, net->metabest will be updated.
 * Basic assumption- the rhs is ALWAYS a base node in all calls here
 *
 */
 
void find_match(net_type *net, int node, nodelist_type *input, int posn)
{
  int metanode,rhs;
  int result_posn;
  int bot,mid,top, safe_mid;
  int explore;
  
  /* printf("*** in find_match, node='"); print_node(net,node); printf("'\n");*/
  
  /* printf("*** NODE HAS %i LHS-OUTPUTS and %i RHS-OUTPUTS\n",
     net->data[node+NUM_LHS_OP],net->data[node+NUM_RHS_OP]); */
  /* set up pointers to start/end of the list of outputs of our LHS node*/
  bot=net->data[node+LHS_OP];
  top=bot+net->data[node+NUM_LHS_OP]-1;
  mid=0; result_posn=FALSE;
  /* printf("bot=%i\n,top=%i\n",bot,top); */
  
  while(bot<=top && result_posn==FALSE)
  {
    mid=bot+(top-bot)/2;
    metanode=net->data[mid];
    if (net->trace>=3) { printf("** looking at metanode '");
			 print_node(net,metanode); printf("'\n"); }
    rhs=net->data[metanode+RHS_IP];
    net->result=0;
    result_posn=compare(net, rhs, input, posn); /*try to find a match*/
    if (result_posn>0)
    {
      /* printf("FOUND MATCH!!!\n"); */
      analyse_match(net, metanode, input, result_posn);
      /* pruning? */
      if (net->bestlevel>=get_level(net,node)+PRUNE)  return;
    }
    else if (net->result<0) bot=mid+1;
    else if (net->result>0) top=mid-1;
  }
  /* If we got a match, then */
  /* explore below and above our match for other potential match candidates*/
  if (result_posn>0)  /* i.e. we got a hit... */
  {
    /* firstly, explore below our hit... */
    safe_mid=mid;
    explore=TRUE;  /* i.e. dont explore if we did'nt find a match */
    if (mid==bot) explore=FALSE;
    while (explore)
    {
      mid--;
      metanode=net->data[mid];
      if (net->trace>=3) { printf("looking at metanode '");
			   print_node(net,metanode); printf("'\n"); }
      rhs=net->data[metanode+RHS_IP];    
      result_posn=compare(net, rhs, input, posn); /*try to find a match*/  
      if (result_posn!=FALSE)
      {
        /* printf("FOUND MATCH!!!\n"); */
        analyse_match(net,metanode,input, result_posn);
        /* pruning? */
        if (net->bestlevel>=get_level(net,node)+PRUNE) return;
      }
      else explore=FALSE;
      if (mid==bot) explore=FALSE;
    }
    
    /* now explore above our hit... */
    mid=safe_mid;
    explore=TRUE;
    if (mid==top) explore=FALSE;
    while (explore)
    {
      mid++;
      metanode=net->data[mid];
    if (net->trace>=3) { printf("looking at metanode '");
			 print_node(net,metanode); printf("'\n"); }
      rhs=net->data[metanode+RHS_IP];    
      result_posn=compare(net, rhs, input, posn); /*try to find a match*/    
      if (result_posn!=FALSE)
      {
        /* printf("FOUND MATCH!!!\n"); */
        analyse_match(net,metanode, input, result_posn);
        /* pruning? */
        if (net->bestlevel>=get_level(net,node)+PRUNE) return;
      }
      else explore=FALSE;
      if (mid==top) explore=FALSE;
    }
  }
  /* lastly, explore options where node is the RHS of a metanode... */
  /* chat-mode only! (waaaay too slow and nondeterministic otherwise)*/
  if (net->mode!=1)
  {
     /* printf("** EXPLORING RHS-OUTPUT OPTIONS NOW\n"); */
     top=net->data[node+NUM_RHS_OP]; bot=net->data[node+RHS_OP]; mid=0;
     while(mid<top)
     {
       metanode=net->data[bot+mid];
       /*printf("RHS-EXPLORE OF '"); print_node(net,metanode); printf("'\n");*/
       find_match(net,metanode,input,posn);
       mid++;
     }
  }
}

/**************************************************************************
 * netlisten() - take in an ascii stream and feed it to the meme machine.
 * mode=1 (listen mode) or 2 (chat mode).
 */
void netlisten(net_type *net, nodelist_type  *input, nodelist_type *output,
	       int tweak)
{
  int previous, posn;
  int finished;
  /*
  int first, loop;
  clock_t time,btime, match_time, build_time, dream_time;
  struct tms tbuf;
  int tweak_finished;
  */
  FILE *fp_out;
  tweaks_type *tweaks;
  char filename[STR_LEN];
  
  tweaks=(tweaks_type *)malloc(sizeof(tweaks_type));
  finished=FALSE;
  snprintf(filename,STR_LEN,"%s/.meme_output.txt",home);
  if (!(fp_out=fopen(filename, "w"))) {
	fprintf(stderr,"fopen('%s')",filename);
	perror("");
	exit(1);
  }
  previous=0; posn=1; net->spoken=0;
  
  /* right, main loop. Chunk the entire input stream. */
  while(!finished)   /*main loop*/
  {
    if (net->trace>=2 && net->mode==1)
      
    printf("-----------------------------------------------------------\n");
    if (posn>input->size && net->mode>1)
    {
      need_input(net,input);
    }
    
    net->start=posn; net->size=input->size;
    net->best=input->data[posn]; net->bestlevel=1; net->bestnext=posn+1;
    net->tweaked=0; net->besttweaked=0; net->best_wmsize=0;
    
    if (tweak==TRUE)
                multi_tweak_match(net,net->best,input,net->bestnext,tweaks);
           else find_match(net,net->best,input, net->bestnext);
    
    /* OK, we've got the best lhs and rhs, now create a higher level */ 
    
    if (previous!=0 && net->best!=0)
    {
       /* printf("**** new node - '"); print_node(net,previous);
          printf("' , '"); print_node(net,thread->best); printf("'\n");*/
       /* Dont build a node where lhs is a responce,but rhs is'nt. */
       if (net->data[previous]>=HIGH_BIT &&
           net->data[net->best]<HIGH_BIT)
       {
           /*dont make a node*/
       }
       else 
       {
         make_node(net,previous,net->best);
       }
    }

    /*print the output to screen and memory... */
    if (net->trace>=1)
    {
      if (net->mode==1)
      {
        print_node(net,net->best); printf("\n"); 
        if (net->trace>=2 && net->besttweaked==1)
         {printf("template='"); print_node(net,net->template); printf("'\n");}
      }
      output->size++; output->data[output->size]=net->best;
    }
    print_node_to_file(net,net->best,fp_out); fprintf(fp_out,"\n");
    previous=net->best;
    /* see if it's time to have a dream.... */
    if (net->actual_num_nodes>net->next_dream)
    { dream(net); previous=0; }
    if (((MAX_DATA_SIZE-net->ptr)+net->nodespace)>RAM)
    {
       if (net->trace>=1)
         printf("**** SHORT ON SPACE! Dream + Defrag forced. ****\n");
       dream(net); defrag(net); previous=0;
    }
    posn=net->bestnext;
    if (net->mode==1 && posn>=input->size) finished=TRUE;
  }
  fclose(fp_out);
}

/**************************************************************************
 * main()
 *  bring net and input text into memory, results of parsing to stdout.
 */

int main(int argc, char *argv[])
{
  net_type *net;
  int fp;
  nodelist_type *input, *output;
  int tweak;  /*whether to use tweaking or non-tweaking method for learning*/
  int arg;  /* for parsing program arguments*/
  char filename[STR_LEN]="input.txt";
  FILE *fp_pid;
  int cycles,loop,fd;
  char pidfile[STR_LEN];
  char err_str[STR_LEN];
  
  
  fprintf(stderr,
	  "\n------------------------------------------------------------\n");
  fprintf(stderr,
	  "Meme Machine - by Brian Smith (brian@wintermute.demon.co.uk)\n"); 
  fprintf(stderr,
	  "               Additonal code by Kees Cook <cook@cpoint.net>\n");

  /* start up some globals */
#ifdef DOS
  chatlog=strdup("/tmp/chatlog");
#else
  home=getenv("HOME");
  snprintf(pidfile,STR_LEN,"%s/.meme_chat.log",home);
  chatlog=strdup(pidfile);
#endif
  if (!chatlog) {
	perror("strdup");
	exit(1);
  }

  /* firstly, write the process PID out... */
#ifdef DOS
#else
  snprintf(pidfile,STR_LEN,"%s/.meme_pid",home);

  if ((fd=open(pidfile,O_RDWR|O_CREAT|O_TRUNC,
	       S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))<0) {
	snprintf(err_str,STR_LEN,"open('%s')",pidfile);
	perror(err_str);
	exit(1);
  }
  if (!(fp_pid=fdopen(fd,"w"))) {
	snprintf(err_str,STR_LEN,"fdopen(%d)",fd);
	perror(err_str);
	exit(1);
  }
  fprintf(fp_pid,"%i\n",getpid());
  fclose(fp_pid); 
#endif

  /* now, create an initial node-net if none exists */
  if  (! fopen("net","rb"))
  {
    printf("creating a blank 'net' file...\n");
    initialise();
  }
  input=(nodelist_type *)malloc(sizeof(nodelist_type)); input->size=0;
  output=(nodelist_type *)malloc(sizeof(nodelist_type)); output->size=0;
  
#ifdef DOS
   net=(net_type *)malloc(sizeof(net_type));
   load(net);
#else
 if ((fp=open("net",O_RDWR))<0) {
	perror("open('net')");
	exit(1);
 }
 net=(net_type *)mmap(0,sizeof(net_type),PROT_READ|PROT_WRITE,MAP_SHARED,fp,0);
 if (!net) {
	perror("mmap('net')");
	exit(1);
 }
#endif

  /* sanity check... */
  if (net->nodespace==0) {printf("WOA! problems with net size.\n"); exit(0);}
  net->mode=1; net->chatlog=0; net->trace=0; tweak=FALSE;
  cycles=1;
  /* now parse the various arguments... */
  
  if (argc==1 && !fopen("input.txt","rb")) /* if no args given, explain usage.*/
  {
    printf("To watch the program learn from a file 'alice.txt', type:\n");
    printf("\nmeme alice.txt -trace 1\n\n");
    printf("The output goes to the screen and a file 'output.txt'.\n");
    printf("To train a network with the text file n times, type\n");
    printf("\nmeme alice.txt -cycles n\n\n");
    printf("To use the chat mode after some training, type:\n");
    printf("\nmeme -chat\n\n");
    printf("To experiment with tweak-based learning, add the option\n");
    printf("'-tweak' To any of the above, but expect a tenfold slowdown.\n");
    printf("---------------------------------------------------------------\n");
    
    exit(0);
  }
  for (arg=1; arg<argc; arg++)
  {
         if (strcmp(argv[arg],"-notweak")==0) tweak=FALSE;
    else if (strcmp(argv[arg],"-tweak")==0) tweak=TRUE;
    else if (strcmp(argv[arg],"-chat")==0) { net->mode=2; tweak=TRUE; }
    else if (strcmp(argv[arg],"-trace")==0) {net->trace=*argv[arg+1]-'0';arg++;}
    else if (strcmp(argv[arg],"-cycles")==0) {cycles=atoi(argv[arg+1]); arg++;}
    else if (strcmp(argv[arg],"-cycle")==0) {cycles=atoi(argv[arg+1]); arg++;}
    else if (strcmp(argv[arg],"-chatlog")==0) net->chatlog=1;
    else if (strcmp(argv[arg],"-defrag")==0) net->mode=42;  /*defrag mode*/
    else if (strcmp(argv[arg],"-stats")==0) {dump_stats(net); return(0);}
    else if (strcmp(argv[arg],"-story")==0) net->mode=69; /*storytext mode*/
    else if (strcmp(argv[arg],"-help")==0 || strcmp(argv[arg],"-?")==0)
    {
       printf("Options:\n-chat - run the program in interactive chat-mode\n");
       printf("-notweak (default) - parse text in no-tweak mode\n");
       printf("-tweak - parse text in tweak mode (slower) \n");
       printf("-chatlog - assume that the text file is a transcript \n");
       printf("-trace X - show what the program is doing. X=1,2,3 or 4)\n");
       printf("-defrag - defragment the 'net' file (can take a long time)\n");
       printf("-stats - show some statistics on the net file\n");
    }
    else if (argv[arg][0]!='-') strncpy(filename, argv[arg],STR_LEN); /* got filename*/
    else { printf("unknown argument: '%s'\n",argv[arg]); exit(0); }
  }

  if (net->mode==42) { defrag(net); return(0);}
  if (net->mode==1 || net->mode==69)    /* we're in training mode.. */
  {
    read_text(net, input,output, filename);
    for(loop=1; loop<=cycles; loop++)
    {
      printf("-----------------------\nTRAINING CYCLE %i\n", loop);
      netlisten(net,input,output,tweak);
    }
    printf("after training, nodespace=%i, dataspace=%i, free=%i \n",
          net->nodespace, MAX_DATA_SIZE-net->ptr, net->ptr-net->nodespace);
  }
  else /* We're in 'chat' mode... */
  {
     input->data=(int *)malloc(100000);
     output->data=(int *)malloc(100000);
     netlisten(net,input,output,tweak);
  }
#ifdef DOS
 printf("writing back to disk...\n");
 save(net);
 printf("finished.\n");
#else
  /* write changed data back */
  if (msync((void *)net,sizeof(net_type), MS_SYNC)) {
	perror("msync('net')");
	exit(1);
  }
  printf("finished.\n");
  close(fp); 
#endif
  return(0);
}
